﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using MySql.Data.MySqlClient;

namespace MontoyaTech.MySqlPlus
{
    /// <summary>
    /// The outline of a managed MySqlConnection that adds extra functionality like automatic
    /// query retrying and reconnecting if the remote DB goes away or becomes busy.
    /// </summary>
    public class MySqlManagedConnection : IDisposable
    {
        /// <summary>
        /// The Internal MySqlConnection for this managed copy.
        /// </summary>
        public MySqlConnection MySqlConnection = null;

        /// <summary>
        /// The connection string that is used to repoen the connection.
        /// </summary>
        private string ConnectionString = null;

        /// <summary>
        /// Wether or not we are connected to the remote server.
        /// </summary>
        public bool IsConnected
        {
            get
            {
                try
                {
                    if (this.MySqlConnection == null || this.MySqlConnection.State != System.Data.ConnectionState.Open || !this.MySqlConnection.Ping())
                        return false;
                }
                catch
                {
                    return false;
                }

                return true;
            }
        }

        /// <summary>
        /// Creates a new managed MySqlConnection and sets its internal
        /// MySql connection.
        /// </summary>
        /// <param name="conn"></param>
        public MySqlManagedConnection(MySqlConnection conn)
        {
            if (string.IsNullOrWhiteSpace(conn.ConnectionString))
                throw new Exception("Connection string must be set on MySqlConnection.");

            this.MySqlConnection = conn;
            this.ConnectionString = conn.ConnectionString;
        }

        /// <summary>
        /// Creates a new managed MySqlConnection and setups up the internal connection string and gets everything ready.
        /// </summary>
        /// <param name="conn"></param>
        public MySqlManagedConnection(string conn)
        {
            if (string.IsNullOrWhiteSpace(conn))
                throw new Exception("Invalid connection string passed, it must not be null or empty.");

            this.MySqlConnection = new MySqlConnection(conn);
            this.ConnectionString = conn;
        }

        /// <summary>
        /// Attempts to reconnect to the server and retrys if needed.
        /// </summary>
        public void Reconnect(int maxRetrys = 5, bool exponentialBackoff = true, bool retry = true)
        {
            int backoffSleep = 2000;
            for (int i = 0; i < maxRetrys; i++)
            {
                try
                {
                    if (this.MySqlConnection == null)
                    {
                        this.MySqlConnection = new MySqlConnection(this.ConnectionString);
                    }
                    else
                    {
                        try { this.MySqlConnection.Close(); } catch { }

                        try { this.MySqlConnection.Dispose(); } catch { }

                        this.MySqlConnection = new MySqlConnection(this.ConnectionString);
                    }

                    this.MySqlConnection.Open();

                    //If we are now connected then stop trying.
                    if (this.IsConnected)
                        break;
                    else
                        throw new Exception("Failed to open a valid MySqlConnection.");
                }
                catch (Exception ex)
                {
                    //See if we have reached our max retry.
                    if (i + 1 >= maxRetrys || !retry)
                        throw;

                    //If not backoff for a bit and see if the server comes back online.
                    if (exponentialBackoff)
                    {
                        Thread.Sleep(backoffSleep);
                        backoffSleep *= 2;
                    }
                }
            }
        }

        /// <summary>
        /// Attempts to connect to the remote MySql server.
        /// </summary>
        public void Connect()
        {
            //Just invoke reconnect it handles everything for us.
            this.Reconnect();
        }

        /// <summary>
        /// Disconnects from the remote MySql server.
        /// </summary>
        public void Disconnect()
        {
            if (this.MySqlConnection != null)
            {
                try { this.MySqlConnection.Close(); } catch { }

                try { this.MySqlConnection.Dispose(); } catch { }

                this.MySqlConnection = null;
            }
        }

        /// <summary>
        /// Executes a data reader for a command and retrys 
        /// </summary>
        /// <param name="command"></param>
        /// <param name="maxRetrys"></param>
        /// <param name="exponentialBackoff"></param>
        /// <param name="retry"></param>
        /// <returns></returns>
        public MySqlDataReader ExecuteReader(MySqlCommand command, int maxRetrys = 5, bool exponentialBackoff = true, bool retry = true)
        {
            if (string.IsNullOrWhiteSpace(command.CommandText))
                throw new ArgumentException("Given command must have CommandText set.");

            int backoffSleep = 2000;
            command.CommandTimeout = 60; //Time in seconds

            for (int i = 0; i < maxRetrys; i++)
            {
                ulong startTimestamp = GlobalTimeStamp.Current;

                try
                {
                    command.Connection = this.MySqlConnection;
                    return command.ExecuteReader();
                }
                catch (Exception ex)
                {
                    //Get the inner most exception.
                    var innerException = ex;
                    while (innerException.InnerException != null)
                        innerException = innerException.InnerException;

                    //Try to get the MySql error code
                    int code = -1;
                    if (ex is MySqlException)
                        code = ((MySqlException)ex).Number;

                    //See if we have reached our max retry.
                    if (i + 1 >= maxRetrys || !retry)
                        throw;

                    //See if the connection was invalid
                    if (innerException.GetType().IsAssignableFrom(typeof(System.Net.Sockets.SocketException)) ||
                        innerException.GetType().IsAssignableFrom(typeof(MySqlConnection)) ||
                        ex.GetType().IsAssignableFrom(typeof(MySqlConnection)) ||
                        ex.Message.ToLower().Contains("connection must be valid and open") ||
                        ex.Message.ToLower().Contains("a connection attempt failed because the connected party did not properly respond after a period of time") ||
                        ex.Message.ToLower().Contains("an existing connection was forcibly closed by the remote host"))
                    {
                        //Attempt to reopen the connection.
                        this.Reconnect(maxRetrys, exponentialBackoff);
                    }

                    //See if we should retry or just throw.
                    if (!ShouldRetryBasedOnMySqlErrorNum(code))
                        throw;

                    //If the operation took less than 5 seconds, then sleep. Otherwise continue right away. This is to prevent OpenRetry from making us wait even longer.
                    if (GlobalTimeStamp.GetDifference(startTimestamp) <= 5000 && exponentialBackoff)
                    {
                        Thread.Sleep(backoffSleep);
                        backoffSleep *= 2;
                    }
                }
            }

            throw new Exception($"MySql ExecuteReader failed. Max timeout reached. Query: {command.CommandText}");
        }

        /// <summary>
        /// Executes a non query command.
        /// </summary>
        /// <param name="command"></param>
        /// <param name="maxRetrys"></param>
        /// <param name="exponentialBackoff"></param>
        /// <param name="retry"></param>
        /// <returns>The number of rows affected.</returns>
        /// <exception cref="Exception"></exception>
        public int ExecuteNonQuery(MySqlCommand command, int maxRetrys = 5, bool exponentialBackoff = true, bool retry = true)
        {
            if (string.IsNullOrWhiteSpace(command.CommandText))
                throw new ArgumentException("Given command must have CommandText set.");

            int backoffSleep = 2000;
            command.CommandTimeout = 60; // time in seconds

            for (int i = 0; i < maxRetrys; i++)
            {
                ulong startTimestamp = GlobalTimeStamp.Current;

                try
                {
                    command.Connection = this.MySqlConnection;

                    return command.ExecuteNonQuery();
                }
                catch (Exception ex)
                {
                    //Get the inner most exception.
                    var innerException = ex;
                    while (innerException.InnerException != null)
                        innerException = innerException.InnerException;

                    //Try to get the MySql error code if we can
                    int code = -1;
                    if (ex is MySqlException)
                        code = ((MySqlException)ex).Number;

                    //See if we have reached our max retry.
                    if (i + 1 >= maxRetrys || !retry)
                        throw;

                    //See if the connection was invalid
                    if (innerException.GetType().IsAssignableFrom(typeof(System.Net.Sockets.SocketException)) ||
                        innerException.GetType().IsAssignableFrom(typeof(MySqlConnection)) ||
                        ex.GetType().IsAssignableFrom(typeof(MySqlConnection)) ||
                        ex.Message.ToLower().Contains("connection must be valid and open") ||
                        ex.Message.ToLower().Contains("a connection attempt failed because the connected party did not properly respond after a period of time") ||
                        ex.Message.ToLower().Contains("an existing connection was forcibly closed by the remote host"))
                    {
                        this.Reconnect(maxRetrys, exponentialBackoff);
                    }

                    //See if we should retry or just throw.
                    if (!ShouldRetryBasedOnMySqlErrorNum(code))
                        throw;

                    //If the operation took less than 5 seconds, then sleep. Otherwise continue right away.
                    if (GlobalTimeStamp.GetDifference(startTimestamp) <= 5000 && exponentialBackoff)
                    {
                        Thread.Sleep(backoffSleep);
                        backoffSleep *= 2;
                    }
                }
            }

            throw new Exception($"MySql ExecuteNonQuery failed. Max timeout reached. Query: {command.CommandText}");
        }

        /// <summary>
        /// Returns whether or not we should retry a query based on the MySql error number.
        /// </summary>
        /// <param name="number"></param>
        /// <returns></returns>
        private static bool ShouldRetryBasedOnMySqlErrorNum(int number)
        {
            //List of codes here: https://www.briandunning.com/error-codes/?source=MySQL
            if (number >= 1044 && number <= 1052)
                return false;
            else if (number >= 1054 && number <= 1075)
                return false;
            else if (number == 1265)
                return false; //Data truncation (Don't try again)
            else if (number == 1264)
                return false; //Value out of range (Don't try again)

            return true;
        }

        /// <summary>
        /// Allows the exposal of the internal MySqlConnection from the managed version.
        /// </summary>
        /// <param name="connection"></param>
        public static implicit operator MySqlConnection(MySqlManagedConnection connection)
        {
            return connection.MySqlConnection;
        }

        /// <summary>
        /// Disposes this managed connection and releases all resources.
        /// </summary>
        public void Dispose()
        {
            if (this.MySqlConnection != null)
            {
                try { this.MySqlConnection.Close(); } catch { }

                try { this.MySqlConnection.Dispose(); } catch { }

                this.MySqlConnection = null;
            }
        }

        /// <summary>
        /// Cleans up this MySqlManagedConnection when GC is collecting.
        /// </summary>
        ~MySqlManagedConnection()
        {
            Dispose();
        }
    }
}
